@@ -10,14 +10,16 @@
 # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
 # ANY KIND, either express or implied. See the License for the specific
 # language governing permissions and limitations under the License.
-"""This module contains the inteface for controlling how configuration
+"""This module contains the interface for controlling how configuration
 is loaded.
 """
+
 import copy
 import logging
 import os
 
 from botocore import utils
+from botocore.exceptions import InvalidConfigError
 
 logger = logging.getLogger(__name__)
 
@@ -46,6 +48,8 @@ logger = logging.getLogger(__name__)
 #: the ``env var`` is the OS environment variable (``os.environ``) to
 #: use, and ``default_value`` is the value to use if no value is otherwise
 #: found.
+#: NOTE: Fixing the spelling of this variable would be a breaking change.
+#: Please leave as is.
 BOTOCORE_DEFAUT_SESSION_VARIABLES = {
     # logical:  config_file, env_var,        default_value, conversion_func
     'profile': (None, ['AWS_DEFAULT_PROFILE', 'AWS_PROFILE'], None, None),
@@ -54,94 +58,166 @@ BOTOCORE_DEFAUT_SESSION_VARIABLES = {
     'config_file': (None, 'AWS_CONFIG_FILE', '~/.aws/config', None),
     'ca_bundle': ('ca_bundle', 'AWS_CA_BUNDLE', None, None),
     'api_versions': ('api_versions', None, {}, None),
-
     # This is the shared credentials file amongst sdks.
-    'credentials_file': (None, 'AWS_SHARED_CREDENTIALS_FILE',
-                         '~/.aws/credentials', None),
-
+    'credentials_file': (
+        None,
+        'AWS_SHARED_CREDENTIALS_FILE',
+        '~/.aws/credentials',
+        None,
+    ),
     # These variables only exist in the config file.
-
     # This is the number of seconds until we time out a request to
     # the instance metadata service.
     'metadata_service_timeout': (
         'metadata_service_timeout',
-        'AWS_METADATA_SERVICE_TIMEOUT', 1, int),
+        'AWS_METADATA_SERVICE_TIMEOUT',
+        1,
+        int,
+    ),
     # This is the number of request attempts we make until we give
     # up trying to retrieve data from the instance metadata service.
     'metadata_service_num_attempts': (
         'metadata_service_num_attempts',
-        'AWS_METADATA_SERVICE_NUM_ATTEMPTS', 1, int),
+        'AWS_METADATA_SERVICE_NUM_ATTEMPTS',
+        1,
+        int,
+    ),
     'ec2_metadata_service_endpoint': (
         'ec2_metadata_service_endpoint',
         'AWS_EC2_METADATA_SERVICE_ENDPOINT',
-        None, None),
+        None,
+        None,
+    ),
     'ec2_metadata_service_endpoint_mode': (
         'ec2_metadata_service_endpoint_mode',
         'AWS_EC2_METADATA_SERVICE_ENDPOINT_MODE',
-        None, None),
+        None,
+        None,
+    ),
+    'ec2_metadata_v1_disabled': (
+        'ec2_metadata_v1_disabled',
+        'AWS_EC2_METADATA_V1_DISABLED',
+        False,
+        utils.ensure_boolean,
+    ),
     'imds_use_ipv6': (
         'imds_use_ipv6',
         'AWS_IMDS_USE_IPV6',
-        False, utils.ensure_boolean),
+        False,
+        utils.ensure_boolean,
+    ),
     'use_dualstack_endpoint': (
         'use_dualstack_endpoint',
         'AWS_USE_DUALSTACK_ENDPOINT',
-        None, utils.ensure_boolean),
+        None,
+        utils.ensure_boolean,
+    ),
     'use_fips_endpoint': (
         'use_fips_endpoint',
         'AWS_USE_FIPS_ENDPOINT',
-        None, utils.ensure_boolean),
+        None,
+        utils.ensure_boolean,
+    ),
+    'ignore_configured_endpoint_urls': (
+        'ignore_configured_endpoint_urls',
+        'AWS_IGNORE_CONFIGURED_ENDPOINT_URLS',
+        None,
+        utils.ensure_boolean,
+    ),
     'parameter_validation': ('parameter_validation', None, True, None),
     # Client side monitoring configurations.
     # Note: These configurations are considered internal to botocore.
     # Do not use them until publicly documented.
     'csm_enabled': (
-        'csm_enabled', 'AWS_CSM_ENABLED', False, utils.ensure_boolean),
+        'csm_enabled',
+        'AWS_CSM_ENABLED',
+        False,
+        utils.ensure_boolean,
+    ),
     'csm_host': ('csm_host', 'AWS_CSM_HOST', '127.0.0.1', None),
     'csm_port': ('csm_port', 'AWS_CSM_PORT', 31000, int),
     'csm_client_id': ('csm_client_id', 'AWS_CSM_CLIENT_ID', '', None),
     # Endpoint discovery configuration
     'endpoint_discovery_enabled': (
-        'endpoint_discovery_enabled', 'AWS_ENDPOINT_DISCOVERY_ENABLED',
-        'auto', None),
+        'endpoint_discovery_enabled',
+        'AWS_ENDPOINT_DISCOVERY_ENABLED',
+        'auto',
+        None,
+    ),
     'sts_regional_endpoints': (
-        'sts_regional_endpoints', 'AWS_STS_REGIONAL_ENDPOINTS', 'legacy',
-        None
+        'sts_regional_endpoints',
+        'AWS_STS_REGIONAL_ENDPOINTS',
+        'legacy',
+        None,
     ),
     'retry_mode': ('retry_mode', 'AWS_RETRY_MODE', 'legacy', None),
     'defaults_mode': ('defaults_mode', 'AWS_DEFAULTS_MODE', 'legacy', None),
     # We can't have a default here for v1 because we need to defer to
     # whatever the defaults are in _retry.json.
     'max_attempts': ('max_attempts', 'AWS_MAX_ATTEMPTS', None, int),
+    'user_agent_appid': ('sdk_ua_app_id', 'AWS_SDK_UA_APP_ID', None, None),
+    'request_min_compression_size_bytes': (
+        'request_min_compression_size_bytes',
+        'AWS_REQUEST_MIN_COMPRESSION_SIZE_BYTES',
+        10240,
+        None,
+    ),
+    'disable_request_compression': (
+        'disable_request_compression',
+        'AWS_DISABLE_REQUEST_COMPRESSION',
+        False,
+        utils.ensure_boolean,
+    ),
+    'sigv4a_signing_region_set': (
+        'sigv4a_signing_region_set',
+        'AWS_SIGV4A_SIGNING_REGION_SET',
+        None,
+        None,
+    ),
 }
 # A mapping for the s3 specific configuration vars. These are the configuration
 # vars that typically go in the s3 section of the config file. This mapping
 # follows the same schema as the previous session variable mapping.
 DEFAULT_S3_CONFIG_VARS = {
-    'addressing_style': (
-        ('s3', 'addressing_style'), None, None, None),
+    'addressing_style': (('s3', 'addressing_style'), None, None, None),
     'use_accelerate_endpoint': (
-        ('s3', 'use_accelerate_endpoint'), None, None, utils.ensure_boolean
+        ('s3', 'use_accelerate_endpoint'),
+        None,
+        None,
+        utils.ensure_boolean,
     ),
     'use_dualstack_endpoint': (
-        ('s3', 'use_dualstack_endpoint'), None, None, utils.ensure_boolean
+        ('s3', 'use_dualstack_endpoint'),
+        None,
+        None,
+        utils.ensure_boolean,
     ),
     'payload_signing_enabled': (
-        ('s3', 'payload_signing_enabled'), None, None, utils.ensure_boolean
+        ('s3', 'payload_signing_enabled'),
+        None,
+        None,
+        utils.ensure_boolean,
     ),
     'use_arn_region': (
-        ['s3_use_arn_region',
-         ('s3', 'use_arn_region')],
-        'AWS_S3_USE_ARN_REGION', None, utils.ensure_boolean
+        ['s3_use_arn_region', ('s3', 'use_arn_region')],
+        'AWS_S3_USE_ARN_REGION',
+        None,
+        utils.ensure_boolean,
     ),
     'us_east_1_regional_endpoint': (
-        ['s3_us_east_1_regional_endpoint',
-         ('s3', 'us_east_1_regional_endpoint')],
-        'AWS_S3_US_EAST_1_REGIONAL_ENDPOINT', None, None
+        [
+            's3_us_east_1_regional_endpoint',
+            ('s3', 'us_east_1_regional_endpoint'),
+        ],
+        'AWS_S3_US_EAST_1_REGIONAL_ENDPOINT',
+        None,
+        None,
     ),
     's3_disable_multiregion_access_points': (
         ('s3', 's3_disable_multiregion_access_points'),
-        'AWS_S3_DISABLE_MULTIREGION_ACCESS_POINTS', None, utils.ensure_boolean
+        'AWS_S3_DISABLE_MULTIREGION_ACCESS_POINTS',
+        None,
+        utils.ensure_boolean,
     ),
 }
 # A mapping for the proxy specific configuration vars. These are
@@ -151,21 +227,30 @@ DEFAULT_PROXIES_CONFIG_VARS = {
     'proxy_ca_bundle': ('proxy_ca_bundle', None, None, None),
     'proxy_client_cert': ('proxy_client_cert', None, None, None),
     'proxy_use_forwarding_for_https': (
-        'proxy_use_forwarding_for_https', None, None, utils.normalize_boolean),
+        'proxy_use_forwarding_for_https',
+        None,
+        None,
+        utils.normalize_boolean,
+    ),
 }
 
 
 def create_botocore_default_config_mapping(session):
     chain_builder = ConfigChainFactory(session=session)
     config_mapping = _create_config_chain_mapping(
-        chain_builder, BOTOCORE_DEFAUT_SESSION_VARIABLES)
+        chain_builder, BOTOCORE_DEFAUT_SESSION_VARIABLES
+    )
     config_mapping['s3'] = SectionConfigProvider(
-        's3', session, _create_config_chain_mapping(
-            chain_builder, DEFAULT_S3_CONFIG_VARS)
+        's3',
+        session,
+        _create_config_chain_mapping(chain_builder, DEFAULT_S3_CONFIG_VARS),
     )
     config_mapping['proxies_config'] = SectionConfigProvider(
-        'proxies_config', session, _create_config_chain_mapping(
-            chain_builder, DEFAULT_PROXIES_CONFIG_VARS)
+        'proxies_config',
+        session,
+        _create_config_chain_mapping(
+            chain_builder, DEFAULT_PROXIES_CONFIG_VARS
+        ),
     )
     return config_mapping
 
@@ -178,13 +263,12 @@ def _create_config_chain_mapping(chain_builder, config_variables):
             env_var_names=config[1],
             config_property_names=config[0],
             default=config[2],
-            conversion_func=config[3]
+            conversion_func=config[3],
         )
     return mapping
 
 
 class DefaultConfigResolver:
-
     def __init__(self, default_config_data):
         self._base_default_config = default_config_data['base']
         self._modes = default_config_data['modes']
@@ -220,13 +304,14 @@ class DefaultConfigResolver:
         return self._resolved_default_configurations[mode]
 
 
-class ConfigChainFactory(object):
+class ConfigChainFactory:
     """Factory class to create our most common configuration chain case.
 
     This is a convenience class to construct configuration chains that follow
     our most common pattern. This is to prevent ordering them incorrectly,
     and to make the config chain construction more readable.
     """
+
     def __init__(self, session, environ=None):
         """Initialize a ConfigChainFactory.
 
@@ -243,9 +328,14 @@ class ConfigChainFactory(object):
             environ = os.environ
         self._environ = environ
 
-    def create_config_chain(self, instance_name=None, env_var_names=None,
-                            config_property_names=None, default=None,
-                            conversion_func=None):
+    def create_config_chain(
+        self,
+        instance_name=None,
+        env_var_names=None,
+        config_property_names=None,
+        default=None,
+        conversion_func=None,
+    ):
         """Build a config chain following the standard botocore pattern.
 
         In botocore most of our config chains follow the the precendence:
@@ -287,8 +377,7 @@ class ConfigChainFactory(object):
         if instance_name is not None:
             providers.append(
                 InstanceVarProvider(
-                    instance_var=instance_name,
-                    session=self._session
+                    instance_var=instance_name, session=self._session
                 )
             )
         if env_var_names is not None:
@@ -329,8 +418,9 @@ class ConfigChainFactory(object):
         return scoped_config_providers
 
 
-class ConfigValueStore(object):
+class ConfigValueStore:
     """The ConfigValueStore object stores configuration values."""
+
     def __init__(self, mapping=None):
         """Initialize a ConfigValueStore.
 
@@ -347,9 +437,18 @@ class ConfigValueStore(object):
                 self.set_config_provider(logical_name, provider)
 
     def __deepcopy__(self, memo):
-        return ConfigValueStore(
-            copy.deepcopy(self._mapping, memo)
-        )
+        config_store = ConfigValueStore(copy.deepcopy(self._mapping, memo))
+        for logical_name, override_value in self._overrides.items():
+            config_store.set_config_variable(logical_name, override_value)
+
+        return config_store
+
+    def __copy__(self):
+        config_store = ConfigValueStore(copy.copy(self._mapping))
+        for logical_name, override_value in self._overrides.items():
+            config_store.set_config_variable(logical_name, override_value)
+
+        return config_store
 
     def get_config_variable(self, logical_name):
         """
@@ -385,7 +484,10 @@ class ConfigValueStore(object):
 
         :returns: configuration provider or None if not defined.
         """
-        if logical_name in self._overrides or logical_name not in self._mapping:
+        if (
+            logical_name in self._overrides
+            or logical_name not in self._mapping
+        ):
             return None
         provider = self._mapping[logical_name]
         return provider
@@ -444,7 +546,6 @@ class ConfigValueStore(object):
 
 
 class SmartDefaultsConfigStoreFactory:
-
     def __init__(self, default_config_resolver, imds_region_provider):
         self._default_config_resolver = default_config_resolver
         self._imds_region_provider = imds_region_provider
@@ -455,8 +556,9 @@ class SmartDefaultsConfigStoreFactory:
     def merge_smart_defaults(self, config_store, mode, region_name):
         if mode == 'auto':
             mode = self.resolve_auto_mode(region_name)
-        default_configs = self._default_config_resolver.get_default_config_values(
-            mode)
+        default_configs = (
+            self._default_config_resolver.get_default_config_values(mode)
+        )
         for config_var in default_configs:
             config_value = default_configs[config_var]
             method = getattr(self, f'_set_{config_var}', None)
@@ -473,8 +575,7 @@ class SmartDefaultsConfigStoreFactory:
                 current_region = self._instance_metadata_region
             else:
                 try:
-                    current_region = \
-                        self._imds_region_provider.provide()
+                    current_region = self._imds_region_provider.provide()
                     self._instance_metadata_region = current_region
                 except Exception:
                     pass
@@ -487,19 +588,28 @@ class SmartDefaultsConfigStoreFactory:
         return 'standard'
 
     def _update_provider(self, config_store, variable, value):
-        provider = config_store.get_config_provider(variable)
+        original_provider = config_store.get_config_provider(variable)
         default_provider = ConstantProvider(value)
-        if isinstance(provider, ChainProvider):
-            provider.set_default_provider(default_provider)
-            return
-        elif isinstance(provider, BaseProvider):
-            default_provider = ChainProvider(providers=[provider, default_provider])
+        if isinstance(original_provider, ChainProvider):
+            chain_provider_copy = copy.deepcopy(original_provider)
+            chain_provider_copy.set_default_provider(default_provider)
+            default_provider = chain_provider_copy
+        elif isinstance(original_provider, BaseProvider):
+            default_provider = ChainProvider(
+                providers=[original_provider, default_provider]
+            )
         config_store.set_config_provider(variable, default_provider)
 
-    def _update_section_provider(self, config_store, section_name, variable,
-                                 value):
-        section_provider = config_store.get_config_provider(section_name)
-        section_provider.set_default_provider(variable, ConstantProvider(value))
+    def _update_section_provider(
+        self, config_store, section_name, variable, value
+    ):
+        section_provider_copy = copy.deepcopy(
+            config_store.get_config_provider(section_name)
+        )
+        section_provider_copy.set_default_provider(
+            variable, ConstantProvider(value)
+        )
+        config_store.set_config_provider(section_name, section_provider_copy)
 
     def _set_retryMode(self, config_store, value):
         self._update_provider(config_store, 'retry_mode', value)
@@ -509,18 +619,20 @@ class SmartDefaultsConfigStoreFactory:
 
     def _set_s3UsEast1RegionalEndpoints(self, config_store, value):
         self._update_section_provider(
-            config_store, 's3', 'us_east_1_regional_endpoint', value)
+            config_store, 's3', 'us_east_1_regional_endpoint', value
+        )
 
     def _set_connectTimeoutInMillis(self, config_store, value):
-        self._update_provider(config_store, 'connect_timeout', value/1000)
+        self._update_provider(config_store, 'connect_timeout', value / 1000)
 
 
-class BaseProvider(object):
+class BaseProvider:
     """Base class for configuration value providers.
 
     A configuration provider has some method of providing a configuration
     value.
     """
+
     def provide(self):
         """Provide a config value."""
         raise NotImplementedError('provide')
@@ -532,6 +644,7 @@ class ChainProvider(BaseProvider):
     Each provider in the chain is called, the first one returning a non-None
     value is then returned.
     """
+
     def __init__(self, providers=None, conversion_func=None):
         """Initalize a ChainProvider.
 
@@ -551,8 +664,7 @@ class ChainProvider(BaseProvider):
 
     def __deepcopy__(self, memo):
         return ChainProvider(
-            copy.deepcopy(self._providers, memo),
-            self._conversion_func
+            copy.deepcopy(self._providers, memo), self._conversion_func
         )
 
     def provide(self):
@@ -569,14 +681,17 @@ class ChainProvider(BaseProvider):
         return None
 
     def set_default_provider(self, default_provider):
-        if self._providers and isinstance(self._providers[-1], ConstantProvider):
+        if self._providers and isinstance(
+            self._providers[-1], ConstantProvider
+        ):
             self._providers[-1] = default_provider
         else:
             self._providers.append(default_provider)
 
-        num_of_constants = sum(isinstance(
-            provider, ConstantProvider
-        ) for provider in self._providers)
+        num_of_constants = sum(
+            isinstance(provider, ConstantProvider)
+            for provider in self._providers
+        )
         if num_of_constants > 1:
             logger.info(
                 'ChainProvider object contains multiple '
@@ -589,11 +704,12 @@ class ChainProvider(BaseProvider):
         return value
 
     def __repr__(self):
-        return '[%s]' % ', '.join([str(p) for p in self._providers])
+        return '[{}]'.format(', '.join([str(p) for p in self._providers]))
 
 
 class InstanceVarProvider(BaseProvider):
     """This class loads config values from the session instance vars."""
+
     def __init__(self, instance_var, session):
         """Initialize InstanceVarProvider.
 
@@ -609,8 +725,7 @@ class InstanceVarProvider(BaseProvider):
 
     def __deepcopy__(self, memo):
         return InstanceVarProvider(
-            copy.deepcopy(self._instance_var, memo),
-            self._session
+            copy.deepcopy(self._instance_var, memo), self._session
         )
 
     def provide(self):
@@ -620,10 +735,7 @@ class InstanceVarProvider(BaseProvider):
         return value
 
     def __repr__(self):
-        return 'InstanceVarProvider(instance_var=%s, session=%s)' % (
-            self._instance_var,
-            self._session,
-        )
+        return f'InstanceVarProvider(instance_var={self._instance_var}, session={self._session})'
 
 
 class ScopedConfigProvider(BaseProvider):
@@ -645,8 +757,7 @@ class ScopedConfigProvider(BaseProvider):
 
     def __deepcopy__(self, memo):
         return ScopedConfigProvider(
-            copy.deepcopy(self._config_var_name, memo),
-            self._session
+            copy.deepcopy(self._config_var_name, memo), self._session
         )
 
     def provide(self):
@@ -660,14 +771,12 @@ class ScopedConfigProvider(BaseProvider):
         return scoped_config.get(self._config_var_name)
 
     def __repr__(self):
-        return 'ScopedConfigProvider(config_var_name=%s, session=%s)' % (
-            self._config_var_name,
-            self._session,
-        )
+        return f'ScopedConfigProvider(config_var_name={self._config_var_name}, session={self._session})'
 
 
 class EnvironmentProvider(BaseProvider):
     """This class loads config values from environment variables."""
+
     def __init__(self, name, env):
         """Initialize with the keys in the dictionary to check.
 
@@ -682,8 +791,7 @@ class EnvironmentProvider(BaseProvider):
 
     def __deepcopy__(self, memo):
         return EnvironmentProvider(
-            copy.deepcopy(self._name, memo),
-            copy.deepcopy(self._env, memo)
+            copy.deepcopy(self._name, memo), copy.deepcopy(self._env, memo)
         )
 
     def provide(self):
@@ -693,7 +801,7 @@ class EnvironmentProvider(BaseProvider):
         return None
 
     def __repr__(self):
-        return 'EnvironmentProvider(name=%s, env=%s)' % (self._name, self._env)
+        return f'EnvironmentProvider(name={self._name}, env={self._env})'
 
 
 class SectionConfigProvider(BaseProvider):
@@ -702,11 +810,13 @@ class SectionConfigProvider(BaseProvider):
     This is useful for retrieving scoped config variables (i.e. s3) that have
     their own set of config variables and resolving logic.
     """
+
     def __init__(self, section_name, session, override_providers=None):
         self._section_name = section_name
         self._session = session
         self._scoped_config_provider = ScopedConfigProvider(
-            self._section_name, self._session)
+            self._section_name, self._session
+        )
         self._override_providers = override_providers
         if self._override_providers is None:
             self._override_providers = {}
@@ -715,15 +825,18 @@ class SectionConfigProvider(BaseProvider):
         return SectionConfigProvider(
             copy.deepcopy(self._section_name, memo),
             self._session,
-            copy.deepcopy(self._override_providers, memo)
+            copy.deepcopy(self._override_providers, memo),
         )
 
     def provide(self):
         section_config = self._scoped_config_provider.provide()
         if section_config and not isinstance(section_config, dict):
-            logger.debug("The %s config key is not a dictionary type, "
-                         "ignoring its value of: %s", self._section_name,
-                         section_config)
+            logger.debug(
+                "The %s config key is not a dictionary type, "
+                "ignoring its value of: %s",
+                self._section_name,
+                section_config,
+            )
             return None
         for section_config_var, provider in self._override_providers.items():
             provider_val = provider.provide()
@@ -739,21 +852,22 @@ class SectionConfigProvider(BaseProvider):
             provider.set_default_provider(default_provider)
             return
         elif isinstance(provider, BaseProvider):
-            default_provider = ChainProvider(providers=[provider, default_provider])
+            default_provider = ChainProvider(
+                providers=[provider, default_provider]
+            )
         self._override_providers[key] = default_provider
 
     def __repr__(self):
         return (
-            'SectionConfigProvider(section_name=%s, '
-            'session=%s, override_providers=%s)' % (
-                self._section_name, self._session,
-                self._override_providers,
-            )
+            f'SectionConfigProvider(section_name={self._section_name}, '
+            f'session={self._session}, '
+            f'override_providers={self._override_providers})'
         )
 
 
 class ConstantProvider(BaseProvider):
     """This provider provides a constant value."""
+
     def __init__(self, value):
         self._value = value
 
@@ -765,4 +879,143 @@ class ConstantProvider(BaseProvider):
         return self._value
 
     def __repr__(self):
-        return 'ConstantProvider(value=%s)' % self._value
+        return f'ConstantProvider(value={self._value})'
+
+
+class ConfiguredEndpointProvider(BaseProvider):
+    """Lookup an endpoint URL from environment variable or shared config file.
+
+    NOTE: This class is considered private and is subject to abrupt breaking
+    changes or removal without prior announcement. Please do not use it
+    directly.
+    """
+
+    _ENDPOINT_URL_LOOKUP_ORDER = [
+        'environment_service',
+        'environment_global',
+        'config_service',
+        'config_global',
+    ]
+
+    def __init__(
+        self,
+        full_config,
+        scoped_config,
+        client_name,
+        environ=None,
+    ):
+        """Initialize a ConfiguredEndpointProviderChain.
+
+        :type full_config: dict
+        :param full_config: This is the dict representing the full
+            configuration file.
+
+        :type scoped_config: dict
+        :param scoped_config: This is the dict representing the configuration
+            for the current profile for the session.
+
+        :type client_name: str
+        :param client_name: The name used to instantiate a client using
+            botocore.session.Session.create_client.
+
+        :type environ: dict
+        :param environ: A mapping to use for environment variables. If this
+            is not provided it will default to use os.environ.
+        """
+        self._full_config = full_config
+        self._scoped_config = scoped_config
+        self._client_name = client_name
+        self._transformed_service_id = self._get_snake_case_service_id(
+            self._client_name
+        )
+        if environ is None:
+            environ = os.environ
+        self._environ = environ
+
+    def provide(self):
+        """Lookup the configured endpoint URL.
+
+        The order is:
+
+        1. The value provided by a service-specific environment variable.
+        2. The value provided by the global endpoint environment variable
+           (AWS_ENDPOINT_URL).
+        3. The value provided by a service-specific parameter from a services
+           definition section in the shared configuration file.
+        4. The value provided by the global parameter from a services
+           definition section in the shared configuration file.
+        """
+        for location in self._ENDPOINT_URL_LOOKUP_ORDER:
+            logger.debug(
+                'Looking for endpoint for %s via: %s',
+                self._client_name,
+                location,
+            )
+
+            endpoint_url = getattr(self, f'_get_endpoint_url_{location}')()
+
+            if endpoint_url:
+                logger.info(
+                    'Found endpoint for %s via: %s.',
+                    self._client_name,
+                    location,
+                )
+                return endpoint_url
+
+        logger.debug('No configured endpoint found.')
+        return None
+
+    def _get_snake_case_service_id(self, client_name):
+        # Get the service ID without loading the service data file, accounting
+        # for any aliases and standardizing the names with hyphens.
+        client_name = utils.SERVICE_NAME_ALIASES.get(client_name, client_name)
+        hyphenized_service_id = (
+            utils.CLIENT_NAME_TO_HYPHENIZED_SERVICE_ID_OVERRIDES.get(
+                client_name, client_name
+            )
+        )
+        return hyphenized_service_id.replace('-', '_')
+
+    def _get_service_env_var_name(self):
+        transformed_service_id_env = self._transformed_service_id.upper()
+        return f'AWS_ENDPOINT_URL_{transformed_service_id_env}'
+
+    def _get_services_config(self):
+        if 'services' not in self._scoped_config:
+            return {}
+
+        section_name = self._scoped_config['services']
+        services_section = self._full_config.get('services', {}).get(
+            section_name
+        )
+
+        if not services_section:
+            error_msg = (
+                f'The profile is configured to use the services '
+                f'section but the "{section_name}" services '
+                f'configuration does not exist.'
+            )
+            raise InvalidConfigError(error_msg=error_msg)
+
+        return services_section
+
+    def _get_endpoint_url_config_service(self):
+        snakecase_service_id = self._transformed_service_id.lower()
+        return (
+            self._get_services_config()
+            .get(snakecase_service_id, {})
+            .get('endpoint_url')
+        )
+
+    def _get_endpoint_url_config_global(self):
+        return self._scoped_config.get('endpoint_url')
+
+    def _get_endpoint_url_environment_service(self):
+        return EnvironmentProvider(
+            name=self._get_service_env_var_name(), env=self._environ
+        ).provide()
+
+    def _get_endpoint_url_environment_global(self):
+        return EnvironmentProvider(
+            name='AWS_ENDPOINT_URL', env=self._environ
+        ).provide()
